@@ -37,12 +37,5 @@ SEXP
     Matrix_statusSym,
     Matrix_uploSym,
     Matrix_xSym,
-    Matrix_zSym,
-    Matrix_REMLSym,
-    Matrix_callSym,
-    Matrix_termsSym,
-    Matrix_assignSym,
-    Matrix_fittedSym,
-    Matrix_residualsSym,
-    Matrix_frameSym;
+    Matrix_zSym;
 
